from .lattice import Value, Bottom, Unknown, PartialTuple, PartialLambda

class _ElemVisitor:
    def is_subseteq_Value(self, other: Value) -> bool: ...
    def is_subseteq_NotConst(self, other: Unknown) -> bool: ...
    def is_subseteq_Unknown(self, other: Bottom) -> bool: ...
    def is_subseteq_PartialTuple(self, other: PartialTuple) -> bool: ...
    def is_subseteq_PartialLambda(self, other: PartialLambda) -> bool: ...
